# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Unit tests for reading/parsing LiteRT-LM files."""

import os
import random

from absl.testing import absltest

from ai_edge_litert.internal import litertlm_core
from ai_edge_quantizer.utils import litertlm_utils
from ai_edge_quantizer.utils import test_utils


class LitertlmUtilsTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    self._litertlm_path = os.path.join(
        test_utils.get_path_to_datafile("../tests/models"),
        "test_tok_tfl_llm.litertlm",
    )

  def test_gets_model_type_from_litertlm(self):
    litertlm_file = litertlm_utils.LiteRTLMFile(self._litertlm_path)
    sections = litertlm_file.sections
    self.assertLen(sections, 4)
    self.assertIsNone(litertlm_file.get_model_type(0))
    self.assertEqual(litertlm_file.get_model_type(1), "tf_lite_prefill_decode")
    self.assertIsNone(litertlm_file.get_model_type(2))
    self.assertIsNone(litertlm_file.get_model_type(3))

  def test_reads_model_from_litertlm(self):
    litertlm_file = litertlm_utils.LiteRTLMFile(self._litertlm_path)
    self.assertIsNotNone(litertlm_file.read_model(1))

  def test_serializes_with_data_overrides(self):
    # Load the original file.
    litertlm_file = litertlm_utils.LiteRTLMFile(self._litertlm_path)

    # Re-serialize the file overriding two of its sections. The section sizes
    # are chosen to _not_ be integer multiples of the LiteRT-LM block size to
    # verify that they are padded correctly.
    output_path = self.create_tempfile("modified.litertlm").full_path
    overrides = {
        0: random.randbytes(10 * litertlm_core.BLOCK_SIZE + 1),
        2: random.randbytes(10 * litertlm_core.BLOCK_SIZE - 1),
    }
    litertlm_file.serialize(output_path, overrides)

    # Read the modified LiteRT-LM file.
    modified_litertlm_file = litertlm_utils.LiteRTLMFile(output_path)

    # Compare the sections.
    self.assertEqual(
        len(litertlm_file.sections), len(modified_litertlm_file.sections)
    )
    for section_id in range(len(litertlm_file.sections)):
      # Check that the key value pairs are the same.
      self.assertEqual(
          litertlm_file.get_section_metadata(section_id),
          modified_litertlm_file.get_section_metadata(section_id),
      )

      # Check whether the buffers match the original or override data.
      if buff := overrides.get(section_id):
        self.assertEqual(
            buff, modified_litertlm_file.get_section_buffer(section_id)
        )
      else:
        self.assertEqual(
            litertlm_file.get_section_buffer(section_id),
            modified_litertlm_file.get_section_buffer(section_id),
        )


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Utilities for QSV update."""

from typing import Any, Union

import numpy as np

from ai_edge_quantizer import qtyping


def _update_moving_average(
    smoothing_factor: Union[np.ndarray, float],
    w: np.ndarray,
    update: np.ndarray,
) -> np.ndarray:
  """Updates weight w with moving average.

  Args:
    smoothing_factor: Smoothing factor used to update w.
    w: Weights to be updated.
    update: Value used for update.

  Returns:
    Weighted sum of w and update.
  """
  return smoothing_factor * w + (1.0 - smoothing_factor) * update


def moving_average_update(
    qsv: qtyping.QSV, new_qsv: qtyping.QSV, smoothing_factor: float = 0.95
) -> qtyping.QSV:
  """Update the QSV (i.e., min/max) using moving average.

  Args:
    qsv: The quantization statistical value of the tensor (min/max) that need to
      be updated.
    new_qsv: The new QSVs (e.g., from new round of calibration).
    smoothing_factor: The weight of moving average.

  Returns:
    The updated QSV for the tensor.
  """
  if not qsv:
    return new_qsv

  updated_qsv = {}
  updated_qsv["min"] = _update_moving_average(
      smoothing_factor, qsv["min"], new_qsv["min"]
  )

  updated_qsv["max"] = _update_moving_average(
      smoothing_factor, qsv["max"], new_qsv["max"]
  )
  return updated_qsv


def _gptq_merge_hessian(
    qsv: qtyping.QSV, new_qsv: qtyping.QSV
) -> tuple[Any, int]:
  """Merges hessian and num_samples from qsv and new_qsv for GPTQ."""
  curr_avg_hessian = qsv["hessian"]
  new_avg_hessian = new_qsv["hessian"]
  curr_samples = qsv["num_samples"]
  new_samples = new_qsv["num_samples"]
  total_samples = curr_samples + new_samples

  if total_samples == 0:
    return new_avg_hessian, 0

  merged_hessian = (
      curr_avg_hessian * curr_samples + new_avg_hessian * new_samples
  ) / total_samples

  return merged_hessian, total_samples


def gptq_and_moving_average_update(
    qsv: qtyping.QSV, new_qsv: qtyping.QSV
) -> qtyping.QSV:
  """Update the QSV with GPTQ logic and moving average logic."""
  if not qsv:
    return new_qsv
  updated_qsv = moving_average_update(qsv, new_qsv)

  merged_hessian, total_samples = _gptq_merge_hessian(qsv, new_qsv)
  updated_qsv["hessian"] = merged_hessian
  updated_qsv["num_samples"] = total_samples
  return updated_qsv


def min_max_update(qsv: qtyping.QSV, new_qsv: qtyping.QSV) -> qtyping.QSV:
  """Update the QSV with minimum min values and maximum max values.

  Args:
    qsv: The quantization statistical value of the tensor (min/max) that need to
      be updated.
    new_qsv: The new QSVs (e.g., from new round of calibration).

  Returns:
    The updated QSV for the tensor.
  """
  if not qsv:
    return new_qsv

  updated_qsv = {}
  updated_qsv["min"] = np.minimum(qsv["min"], new_qsv["min"])
  updated_qsv["max"] = np.maximum(qsv["max"], new_qsv["max"])
  return updated_qsv



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""A command-line tool to quantize TFLite models using a quantization recipe."""

import argparse
from collections.abc import Sequence
import logging
import pathlib
import sys

import os
import io
from ai_edge_quantizer import qtyping
from ai_edge_quantizer import quantizer
from ai_edge_quantizer.utils import litertlm_utils
from ai_edge_quantizer.utils import progress_utils
from ai_edge_quantizer.utils import recipe_utils


def _verify_output_path(
    output_path: str, overwrite_outputs: bool = False
) -> bool:
  """Checks whether the output file exists and whether it's OK to overwrite it."""
  if os.path.exists(output_path) and not overwrite_outputs:
    overwrite_input = input(
        f"Output file {output_path} already exists. Overwrite? (y/N): "
    )
    return overwrite_input.lower() == "y"
  return True


def _quantize_model(
    model: qtyping.Path | qtyping.BufferType,
    recipe: str | qtyping.ModelQuantizationRecipe,
    serialize_to_path: qtyping.Path | None = None,
    enable_progress_report: bool = True,
) -> quantizer.QuantizationResult:
  """Quantizes the given model with the given recipe and maybe write it to disk."""
  qt = quantizer.Quantizer(model)
  qt.load_quantization_recipe(recipe)
  return qt.quantize(
      serialize_to_path=serialize_to_path,
      enable_progress_report=enable_progress_report,
  )


def quantize_litertlm(
    litertlm_path: str,
    recipe: str,
    output_dir: str | None = None,
    overwrite_outputs: bool = False,
) -> int:
  """Quantizes a TFLite model using a recipe.

  Args:
    litertlm_path: Path to the `.litertlm` file containing the models to be
      quantized.
    recipe: Optional recipe name or path to a .json file containing either a
      recipe or a mapping of model_types to quantization recipes.
    output_dir: Optional path to the directory to save the quantized model(s).
      If `None`, the base directory of `litertlm_path` will be used.
    overwrite_outputs: Output files overwrite exisiting files without requiring
      user input.

  Returns:
    `0` on success and a non-zero exit code otherwise.
  """

  litertlm_path = pathlib.Path(litertlm_path)
  litertlm_basename = litertlm_path.stem

  # Load the quantization recipe.
  recipe_map = recipe_utils.resolve_litertlm_recipe_or_mapping(recipe)
  recipe_basename = pathlib.Path(recipe).stem
  default_recipe = recipe_map.get("default")

  # Load the `.litertlm` file.
  litertlm_file = litertlm_utils.LiteRTLMFile(litertlm_path)

  # Create the output directory if it doesn't already exist.
  if not output_dir:
    output_dir = litertlm_path.parent
  else:
    output_dir = pathlib.Path(output_dir)
    if not os.path.exists(output_dir):
      os.makedirs(output_dir)

  # Track progress.
  progress_report = progress_utils.ProgressReport(litertlm_path.name)
  progress_report.capture_progess_start()

  # Loop over the models selected for quantization and populate a dictionary
  # mapping section IDs to serialized quantized models.
  quantized_sections: dict[int, qtyping.BufferType] = {}
  for section_id, section in enumerate(litertlm_file.sections):

    # Skip non-TFLite sections.
    if section.dataType != litertlm_utils.AnySectionDataType.TFLiteModel:
      continue

    # Try to identify the model type.
    if (model_type := litertlm_file.get_model_type(section_id)) is None:
      logging.warning(
          "Could not get the model_type for the TFLiteModel in section %d,"
          " skipping.",
          section_id,
      )
      continue

    # Get the recipe for this model type, skip if none is given.
    if (
        model_recipe := recipe_map.get(model_type, default_recipe)
    ) is None:
      logging.info(
          "No quantization recipe specified for model_type '%s', skipping.",
          model_type,
      )
      continue

    # Create a filename for the quantized TFLite model.
    output_filename = (
        f"{litertlm_basename}_{section_id:03d}_{model_type}_"
        f"{recipe_basename}.tflite"
    )
    output_file_path = str(output_dir / output_filename)
    if not _verify_output_path(output_file_path, overwrite_outputs):
      logging.error("Aborting.")
      return 1

    # Quantize the TFLite model and write it to disk.
    quantized_sections[section_id] = _quantize_model(
        model=litertlm_file.get_section_buffer(section_id),
        recipe=model_recipe,
        serialize_to_path=output_file_path,
        enable_progress_report=False,
    ).quantized_model

  if not quantized_sections:
    logging.error("No models were quantized, not creating output file.")
    return 1

  # Rebuild the LiteRT-LM file with the quantized models swapped in.
  output_filename = f"{litertlm_basename}_{recipe_basename}.litertlm"
  output_file_path = str(output_dir / output_filename)
  if not _verify_output_path(output_file_path, overwrite_outputs):
    logging.error("Aborting.")
    return 1
  output_file_size = litertlm_file.serialize(
      output_file_path, quantized_sections
  )

  progress_report.generate_progress_report(
      os.path.getsize(litertlm_path), output_file_size
  )

  return 0


def quantize_tflite(
    model_file: str,
    recipe_file: str,
    output_dir: str | None = None,
    overwrite_outputs: bool = False,
) -> int:
  """Quantizes a TFLite model using a recipe.

  Args:
    model_file: Path to the .tflite file to be quantized.
    recipe_file: Path to the .json file with the quantization recipe.
    output_dir: Optional path to the directory to save the quantized model(s).
      If `None`, the base directory of `model_file` will be used.
    overwrite_outputs: Output files overwrite exisiting files without requiring
      user input.

  Returns:
    `0` on success and a non-zero exit code otherwise.
  """
  model_file = pathlib.Path(model_file)
  model_basename = model_file.stem
  recipe_basename = pathlib.Path(recipe_file).stem
  output_filename = f"{model_basename}_{recipe_basename}.tflite"

  # Create the output directory if it doesn't already exist.
  if not output_dir:
    output_dir = model_file.parent
  else:
    output_dir = pathlib.Path(output_dir)
    if not os.path.exists(output_dir):
      os.makedirs(output_dir)

  output_file_path = str(output_dir / output_filename)

  if not _verify_output_path(output_file_path, overwrite_outputs):
    logging.error("Aborting.")
    return 1

  _quantize_model(model_file, recipe_file, serialize_to_path=output_file_path)

  return 0


def parse_args(args: Sequence[str]) -> argparse.Namespace:
  """Parses command-line arguments.

  Args:
    args: A list of strings to parse. If None, sys.argv is used.

  Returns:
    An argparse.Namespace containing the parsed arguments,
  """
  parser = argparse.ArgumentParser(
      description="Quantize TFLite models using a quantization recipe.",
  )
  parser.add_argument(
      "--model_file",
      required=True,
      help="Path to the .tflite or .litertlm file to be quantized.",
  )
  parser.add_argument(
      "--recipe",
      required=True,
      help=(
          "Recipe name or path to the .json file with the quantization recipe."
      ),
  )
  parser.add_argument(
      "--output_dir",
      help=(
          "Path to the directory in which to save the quantized model(s). If"
          " not specified, the directory of the `model_file` argument will be"
          " used."
      ),
  )
  parser.add_argument(
      "--overwrite_outputs",
      action="store_true",
      help="Overwrite exisiting output files without requesting user input.",
  )
  return parser.parse_args(args[1:])


def main(parsed_args: argparse.Namespace) -> int:
  if parsed_args.model_file.endswith(".tflite"):
    return quantize_tflite(
        model_file=parsed_args.model_file,
        recipe_file=parsed_args.recipe,
        output_dir=parsed_args.output_dir,
        overwrite_outputs=parsed_args.overwrite_outputs,
    )
  elif parsed_args.model_file.endswith(".litertlm"):
    return quantize_litertlm(
        litertlm_path=parsed_args.model_file,
        recipe=parsed_args.recipe,
        output_dir=parsed_args.output_dir,
        overwrite_outputs=parsed_args.overwrite_outputs,
    )
  logging.error(
      'File passed to `--model_file` must end in either ".tflite" or'
      ' ".litertlm".'
  )
  return 1


# Wrapper to play nicely with uv tool install ai-edge-quantizer
def cli_main():
  main(parse_args(sys.argv))


if __name__ == "__main__":
  cli_main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Utilities to read and parse LiteRT-LM files."""

from collections.abc import Mapping
import copy
import pathlib
import struct
from typing import Any, TypeVar

import flatbuffers

import os
import io
from ai_edge_litert.internal import litertlm_builder
from ai_edge_litert.internal import litertlm_core
from ai_edge_litert.internal import litertlm_header_schema_py_generated as schema
from ai_edge_litert.tools import flatbuffer_utils
from ai_edge_litert.tools import mmap_utils
from ai_edge_quantizer import qtyping

# Exported types.
SectionObjectT = schema.SectionObjectT
AnySectionDataType = schema.AnySectionDataType

# Internal types.
_Path = str | pathlib.Path
T = TypeVar('T')


# TODO: b/495756579 - Fix a mis-named function in the generated schema.
schema.VDataCreator = schema.VdataCreator


def _bytes_to_str(val: bytes | T) -> str | T:
  """Converts the input to `str` if it is `bytes`."""
  return str(val, encoding='utf-8') if isinstance(val, bytes) else val


class LiteRTLMFile:
  """Represents a LiteRT-LM file, which may contain one or more TFLite models.

  On initialization, the LiteRTLM file header is read and parsed, and a list
   sections (offsets in the LiteRTLM file) is created.

  Subsequent calls to `get_section_type` or `read_model` will use this list to
  extract data from the LiteRT-LM file.
  """

  _path: _Path
  _metadata: schema.LiteRTLMMetaDataT
  _sections: list[SectionObjectT]

  def __init__(self, path: _Path):
    self._path = path
    self._metadata = self._get_metadata()
    self._sections = self._get_sections()

  def _get_metadata(self) -> schema.LiteRTLMMetaDataT:
    """Extracts the packed metadata from the file header."""
    # TODO: b/495763732 - This should be part of litertlm_core.

    # Read and check the file header to extract the size of the serializaed
    # metadata.
    header_bytes = mmap_utils.get_file_contents(
        self._path, size=litertlm_core.HEADER_BEGIN_BYTE_OFFSET
    )
    assert header_bytes[:8] == litertlm_core.HEADER_MAGIC_BYTES

    # Get the end offset of the serialized metadata.
    header_end_offset = struct.unpack(
        '<Q',
        header_bytes[litertlm_core.HEADER_END_LOCATION_BYTE_OFFSET :][:8],
    )[0]

    # Extract and parse the metadata flatbuffer.
    metadata_buffer = mmap_utils.get_file_contents(
        self._path,
        offset=litertlm_core.HEADER_BEGIN_BYTE_OFFSET,
        size=header_end_offset - litertlm_core.HEADER_BEGIN_BYTE_OFFSET,
    )
    return schema.LiteRTLMMetaDataT.InitFromPackedBuf(metadata_buffer)

  def _get_sections(self) -> list[SectionObjectT]:
    """Extracts the mapping of `model_type` to `SectionObjectT`."""
    # Get the SectionObjects from the metadata.
    if (
        (metadata := self._metadata)
        and (section_metadata := metadata.sectionMetadata)
        and (objects := section_metadata.objects)
    ):
      return objects[:]
    return []

  @property
  def sections(self) -> list[SectionObjectT]:
    """Returns a list of `SectionObjectT` contained in this LiteRT-LM file."""
    return self._sections

  def get_system_metadata(self) -> dict[str, Any]:
    """Returns a `dict` of key/value pairs corresponding to the system metadata of this LiteRT-LM file."""
    if (system_metadata := self._metadata.systemMetadata) and (
        entries := system_metadata.entries
    ):
      return {
          _bytes_to_str(entry.key): _bytes_to_str(entry.value.value)
          for entry in entries
      }
    return {}

  def get_section_metadata(self, section_id: int) -> dict[str, Any]:
    """Returns a `dict` of key/value pairs corresponding to the system metadata of the given section of the LiteRT-LM file."""
    if items := self._sections[section_id].items:
      return {
          _bytes_to_str(item.key): _bytes_to_str(item.value.value)
          for item in items
      }
    return {}

  def get_model_type(self, section_id: int) -> str | None:
    """Returns the model_type of the given section of the LiteRT-LM file."""
    for key, val in self.get_section_metadata(section_id).items():
      if key == 'model_type':
        return _bytes_to_str(val)
    return None

  def read_model(self, section_id: int) -> qtyping.ModelT | None:
    """Extracts a `qtyping.ModelT` from the given section of this LiteRT-LM file."""
    obj = self._sections[section_id]
    if obj.dataType == AnySectionDataType.TFLiteModel:
      return flatbuffer_utils.read_model_from_bytearray(
          mmap_utils.get_file_contents(
              self._path,
              offset=obj.beginOffset,
              size=obj.endOffset - obj.beginOffset,
          )
      )
    return None

  def get_section_buffer(self, section_id: int) -> qtyping.BufferType | None:
    """Extracts the raw contents of the given section of this LiteRT-LM file."""
    obj = self._sections[section_id]
    return mmap_utils.get_file_contents(
        self._path,
        offset=obj.beginOffset,
        size=obj.endOffset - obj.beginOffset,
    )

  def serialize(
      self,
      path: _Path,
      section_data_overrides: Mapping[int, qtyping.BufferType],
  ) -> int:
    """Serializes the LiteRTLMFile to the given path.

    Args:
      path: Location at which to create the LiteRT-LM file.
      section_data_overrides: An optional mapping of `int` to
        `qtyping.BufferType` where the value will replace the contents of the
        section matching the key.

    Returns:
      The number of bytes written to the specified file.
    """
    # Compute the length and offset of each section. Note that there is no
    # guarantee that the sections are ordered by their location in the
    # LiteRT-LM file!
    section_offsets = [
        min(self._sections, key=lambda s: s.beginOffset).beginOffset
    ]
    section_lengths = []
    for section_id, section in enumerate(self._sections):
      if buff := section_data_overrides.get(section_id):
        section_length = len(buff)
      else:
        section_length = section.endOffset - section.beginOffset
      section_lengths.append(section_length)
      section_offsets.append(
          (section_offsets[-1] + section_length + litertlm_core.BLOCK_SIZE - 1)
          & ~(litertlm_core.BLOCK_SIZE - 1)
      )

    # Make a copy of the metadata with the modified section offsets.
    metadata = copy.deepcopy(self._metadata)
    assert metadata.sectionMetadata and metadata.sectionMetadata.objects
    for section_id, section in enumerate(metadata.sectionMetadata.objects):
      section.beginOffset = section_offsets[section_id]
      section.endOffset = section.beginOffset + section_lengths[section_id]

    # Update the system metadata where needed.
    if not metadata.systemMetadata:
      metadata.systemMetadata = schema.SystemMetadataT(entries=[])
    system_metadata = metadata.systemMetadata
    system_metadata.entries = [
        m.to_key_value_pair()
        for m in litertlm_builder.populate_system_metadata([
            litertlm_builder.Metadata.from_key_value_pair(kvp)
            for kvp in system_metadata.entries
        ])
    ]

    # Pack the modified metadata.
    metadata_builder = flatbuffers.Builder(1024)
    metadata_builder.Finish(metadata.Pack(metadata_builder))
    assert metadata_builder.Offset() <= section_offsets[0]

    # Create a memory-mapped file in which we will pack the LiteRTLM file.
    if (
        buff := mmap_utils.get_mapped_buffer_or_none(path, section_offsets[-1])
    ) is None:
      buff = open(path, 'wb')

    # Write the header and metadata. Note that we don't use `buff.seek` to avoid
    # failing on file systems that don't support it.
    # TODO: b/495763732 - This should be part of litertlm_core.
    buff_offset = 0
    buff_offset += buff.write(litertlm_core.HEADER_MAGIC_BYTES)
    buff_offset += buff.write(
        litertlm_core.LITERTLM_MAJOR_VERSION.to_bytes(4, 'little')
    )
    buff_offset += buff.write(
        litertlm_core.LITERTLM_MINOR_VERSION.to_bytes(4, 'little')
    )
    buff_offset += buff.write(
        litertlm_core.LITERTLM_PATCH_VERSION.to_bytes(4, 'little')
    )
    buff_offset += buff.write(int(0).to_bytes(4, 'little'))  # Zero padding.

    # Write the offset of the end of the serialized metadata flatbuffer.
    assert buff_offset == litertlm_core.HEADER_END_LOCATION_BYTE_OFFSET
    buff_offset += buff.write(
        (
            litertlm_core.HEADER_BEGIN_BYTE_OFFSET + metadata_builder.Offset()
        ).to_bytes(8, 'little')
    )

    # Write the serialized metadata flatbuffer.
    assert buff_offset == litertlm_core.HEADER_BEGIN_BYTE_OFFSET
    buff_offset += buff.write(metadata_builder.Output())

    # Write the section data.
    for section_id, section in enumerate(metadata.sectionMetadata.objects):
      # Move to the aligned offset of the next section.
      buff_offset += buff.write(b'\x00' * (section.beginOffset - buff_offset))

      # Write the section data.
      if (section_data := section_data_overrides.get(section_id)) is None:
        section_data = self.get_section_buffer(section_id)
      buff_offset += buff.write(section_data)

    # Close the output file/mmap.
    buff.close()

    # Return the number of bytes written.
    return buff_offset



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Unit tests for reading/parsing quantization recipes."""

import json
import os
import pathlib

from absl.testing import absltest

import os
import io
from ai_edge_quantizer import qtyping
from ai_edge_quantizer.utils import recipe_utils
from ai_edge_quantizer.utils import test_utils


class RecipeUtilsTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    self._recipe_file_path = (
        pathlib.Path(test_utils.get_path_to_datafile("."))
        / "../recipes/dynamic_wi8_afp32_recipe.json"
    )
    with open(self._recipe_file_path, "r") as f:
      self._recipe = json.load(f)

  def test_resolves_recipe_from_absolute_path(self):
    self.assertCountEqual(
        self._recipe,
        recipe_utils.resolve_recipe(self._recipe_file_path.absolute()),
    )

  def test_resolves_recipe_from_relative_path(self):
    self.assertCountEqual(
        self._recipe,
        recipe_utils.resolve_recipe(
            self._recipe_file_path.relative_to(os.getcwd())
        ),
    )

  def test_resolves_recipe_from_fileame_in_recipe_dir(self):
    self.assertCountEqual(
        self._recipe,
        recipe_utils.resolve_recipe(self._recipe_file_path.name),
    )

  def test_resolves_recipe_from_name(self):
    self.assertCountEqual(
        self._recipe, recipe_utils.resolve_recipe("dynamic_wi8_afp32")
    )

  def test_resolve_litertlm_quant_recipe_files_correctly(self):
    # Load a single quantization recipe.
    with open(self._recipe_file_path, "r") as f:
      orig_recipe: qtyping.ModelQuantizationRecipe = json.load(f)

    # Create a copy of the original recipe and mark it as such.
    copied_recipe = [{"copied": True}] + orig_recipe

    # Create an output directory outside of `cwd` and `recipes/` that will
    # contain the LiteRT-LM recipe mapping along with a recipe file.
    output_dir = pathlib.Path(self.create_tempdir()) / "recipes"
    os.mkdir(output_dir)
    litertlm_recipe_path = output_dir / "generated_litertlm_recipe.json"
    copied_recipe_path = output_dir / "copied_recipe.json"

    # Create a litertlm recipe mapping.
    orig_litertlm_recipe = {
        # The original recipe's name (will load from `recipes/`).
        "name": "dynamic_wi8_afp32",
        # The recipe itself (will parse from JSON).
        "json": [{"json": True}] + orig_recipe,
        # The full path of the copied recipe.
        "abspath": str(copied_recipe_path),
        # The name of the file in the `recipes/` directory,
        "recipes": self._recipe_file_path.name,
        # The name of the file in the mapping file's directory,
        "basename": copied_recipe_path.name,
    }

    # Write the copied recipe and the mapping to disk.
    with open(copied_recipe_path, "w") as f:
      json.dump(copied_recipe, f)
    with open(litertlm_recipe_path, "w") as f:
      json.dump(orig_litertlm_recipe, f)

    # Load the LiteRT-LM recipe mapping.
    self.assertIsNotNone(
        litertlm_recipe_mapping := recipe_utils.resolve_litertlm_recipe_or_mapping(
            litertlm_recipe_path
        )
    )

    # Check that the mapped recipes resolve to the corresponding quantization
    # recipes.
    for model_type, recipe in litertlm_recipe_mapping.items():
      match model_type:
        case "name" | "recipes":
          self.assertCountEqual(orig_recipe, recipe)
        case "json":
          self.assertEqual({"json": True}, recipe[0])
          self.assertCountEqual(orig_recipe, recipe[1:])
        case "abspath" | "basename":
          self.assertEqual({"copied": True}, recipe[0])
          self.assertCountEqual(orig_recipe, recipe[1:])
        case _:
          self.fail(f"Unexpected model_type {model_type}.")


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Quantization Calibration."""

from collections.abc import Iterable, Mapping
import copy
import enum
import json
from typing import Any, Callable, cast, Union

import numpy as np
from typing_extensions import override

import os
import io
from ai_edge_quantizer import algorithm_manager
from ai_edge_quantizer import default_policy as policy
from ai_edge_quantizer import qtyping
from ai_edge_quantizer import recipe
from ai_edge_quantizer import recipe_manager
from ai_edge_quantizer.utils import calibration_utils
from ai_edge_quantizer.utils import progress_utils
from ai_edge_quantizer.utils import qsv_utils
from ai_edge_quantizer.utils import tfl_flatbuffer_utils
from ai_edge_quantizer.utils import tfl_interpreter_utils
from ai_edge_litert import interpreter as tfl  # pylint: disable=g-direct-tensorflow-import


class CalibrationMode(enum.Enum):
  INFERENCE = 1  # Inference only mode. No calibration is performed.
  CALIBRATION_PRESERVE_ALL_TENSORS = 2  # Calibration with XNNPACK enabled that
  # preserves all intermediate tensors.
  CALIBRATION = 2  # Keeping this for backward compatibility. Please use
  # CALIBRATION_PRESERVE_ALL_TENSORS instead of this.


_SignatureInput = dict[str, Any]  # input_argument_name -> tensor_value.
_SignatureOutput = dict[
    str, np.ndarray
]  # output_argument_name -> tensor_value.
_MISSING_FUNC = cast(
    Callable[[qtyping.QSV, qtyping.QSV], qtyping.QSV], object()
)


class CalibrationInterpreter:
  """A TFL interpreter-like interface for model calibration.

  This is a wrapper around Calibrator that replaces the TFL Interpreter to
  enable calibration. If mode is CALIBRATION, it runs calibration, otherwise it
  acts as a regular TFL interpreter for inference. When in CALIBRATION mode,
  each invocation of a signature runner will update the calibration statistics
  in self._calibrator. Calibrator is needed in both modes because it contains
  tfl interpreter instance to run the model.
  """

  def __init__(
      self,
      model_path: str,
      mode: CalibrationMode = CalibrationMode.INFERENCE,
      qsv_update_func: Callable[
          [qtyping.QSV, qtyping.QSV],
          qtyping.QSV,
      ] = _MISSING_FUNC,
  ):
    """Initializes the CalibrationInterpreter.

    Args:
      model_path: The path to the TFLite model.
      mode: The mode of the interpreter. If CALIBRATION, the interpreter will
        preserve all tensors for calibration purposes.
      qsv_update_func: The function to update the QSVs across calibration steps.
    """
    self._calibrator = Calibrator(
        model_path,
        mode=mode,
        qsv_update_func=qsv_update_func,
    )
    self._mode = mode

  def get_signature_runner(self, signature_key: str | None = None):
    """Returns the signature runner."""
    return CalibrationSignatureRunner(
        self._calibrator, signature_key, self._mode
    )

  def get_calibration_results(self):
    """Returns the calibration results."""
    if self._mode == CalibrationMode.INFERENCE:
      raise ValueError(
          "Calibration results are not available in INFERENCE mode."
      )
    return self._calibrator.get_model_qsvs()

  def save_calibration_result(
      self, output_path: str, extra_metadata: Mapping[str, str] | None = None
  ) -> None:
    """Saves the calibration results."""
    if self._mode == CalibrationMode.INFERENCE:
      raise ValueError(
          "Calibration results are not available in INFERENCE mode."
      )
    self._calibrator.save_calibration_result(output_path, extra_metadata)

  def get_signature_list(self) -> list[str]:
    """Returns the signature list."""
    return self._calibrator.get_signature_list()


class CalibrationSignatureRunner:
  """Wrapper around TFL signature runner to enable calibration."""

  def __init__(
      self,
      calibrator_obj: "Calibrator",
      signature_key: str | None = None,
      mode: CalibrationMode = CalibrationMode.INFERENCE,
      quantization_recipe: qtyping.ModelQuantizationRecipe = recipe.static_wi8_ai8(),
  ):
    """Initializes the CalibrationSignatureRunner.

    Args:
      calibrator_obj: The Calibrator instance.
      signature_key: The key of the signature to run. If None, the default
        signature is used.
      mode: The mode of the runner. If CALIBRATION, invoking the runner will
        update 'calibrator_obj' with new quantization statistics values. If
        INFERENCE, the runner behaves like a standard signature runner.
      quantization_recipe: The quantization recipe to use for calibration.
        Defaults to static_wi8_ai8.
    """
    self._calibrator = calibrator_obj
    self._signature_key = signature_key
    self._mode = mode
    self._recipe_manager = recipe_manager.RecipeManager()
    self._recipe_manager.load_quantization_recipe(quantization_recipe)
    self._signature_runner = (
        self._calibrator._tfl_interpreter.get_signature_runner(
            self._signature_key
        )
    )

  def __call__(self, **kwargs):
    if self._mode == CalibrationMode.INFERENCE:
      return self._signature_runner(**kwargs)
    self._calibrator.calibrate(
        calibration_dataset={self._signature_key: [kwargs]},
        model_recipe_manager=self._recipe_manager,
        cache_output=True,
    )
    outputs = self._calibrator.get_cached_output()
    assert len(outputs) == 1
    self._calibrator.clear_cached_output()
    return outputs[0]

  def get_input_details(self):
    """Returns the input details of the model."""
    return self._signature_runner.get_input_details()

  def get_output_details(self):
    """Returns the output details of the model."""
    return self._signature_runner.get_output_details()


class Calibrator:
  """Base class and factory for TFLite model calibrators.

  This class acts both as a factory for instantiating specific algorithm
  subclasses (via __new__) and as the shared interface for them.

  Notes on Design Architecture:

    1. Calibrator is not an abstract class because it uses the factory pattern
    via __new__ to return instances of its subclasses. Making it abstract would
    cause static type checkers (like pytype) to flag direct instantiations
    as errors. Instead, we raise NotImplementedError in methods that must be
    overridden by subclasses (e.g., _create_interpreter, _calibrate_step).

    2. We prohibit direct instantiation of any subclass of Calibrator to enforce
    the use of the factory pattern implemented in __new__. Meaning no custom
    wrapper classes are allowed to inherit directly from Calibrator.
  """

  def __new__(
      cls,
      float_tflite: Union[str, bytes],
      num_threads: int = 16,
      mode: CalibrationMode = CalibrationMode.CALIBRATION_PRESERVE_ALL_TENSORS,
      qsv_update_func: Callable[
          [qtyping.QSV, qtyping.QSV],
          qtyping.QSV,
      ] = _MISSING_FUNC,
  ) -> "Calibrator":
    """Creates a new instance of the appropriate calibrator subclass.

    Args:
      float_tflite: The path to the TFLite model or the model content as bytes.
      num_threads: The number of threads to use for the TFLite interpreter.
      mode: The mode of the calibrator. Check the docstring of CalibrationMode
        for more details.
      qsv_update_func: The function to update the QSVs across calibration steps.

    Returns:
      An instance of the appropriate calibrator subclass based on the mode.

    Raises:
      TypeError: If trying to instantiate a subclass of Calibrator directly.
    """
    if cls is not Calibrator:
      raise TypeError(
          f"Direct instantiation of {cls.__name__} is not allowed. "
          "Please use the Calibrator factory class instead."
      )

    del float_tflite, num_threads, qsv_update_func

    if mode == CalibrationMode.INFERENCE:
      return super().__new__(_InferenceOnlyCalibrator)
    elif mode == CalibrationMode.CALIBRATION_PRESERVE_ALL_TENSORS:
      return super().__new__(_PreserveAllTensorsCalibrator)
    else:
      raise ValueError(f"Unsupported calibration mode: {mode}")

  def __init__(
      self,
      float_tflite: Union[str, bytes],
      num_threads: int = 16,
      mode: CalibrationMode = CalibrationMode.CALIBRATION_PRESERVE_ALL_TENSORS,
      qsv_update_func: Callable[
          [qtyping.QSV, qtyping.QSV],
          qtyping.QSV,
      ] = _MISSING_FUNC,
  ):
    """Initializes the Calibrator. Check details in docstring of __new__."""
    del mode  # Used only in __new__ and passed to __init__ automatically.

    self._flatbuffer_model = tfl_flatbuffer_utils.read_model(float_tflite)
    self._tfl_interpreter = self._create_interpreter(float_tflite, num_threads)
    if qsv_update_func is _MISSING_FUNC:
      self._qsv_update_func = qsv_utils.moving_average_update
      self._is_custom_qsv_update_func = False
    else:
      self._qsv_update_func = qsv_update_func
      self._is_custom_qsv_update_func = True
    # Tensor name to tensor content.
    self._tensor_content_map: dict[str, Any] = {}
    # QSV of all the tensors in the model.
    self._model_qsvs: dict[str, qtyping.QSV] = {}
    # Cached output of the model.
    self._cached_output: list[_SignatureOutput] = []
    # Metadata for the calibration result.
    self._metadata: dict[str, Any] = {"num_samples_calibrated": 0}

  def _create_interpreter(
      self,
      float_tflite: Union[str, bytes],
      num_threads: int,
  ) -> tfl.Interpreter:
    """Creates the TFLite interpreter."""
    raise NotImplementedError(
        "Subclasses must implement _create_interpreter()."
    )

  def _calibrate_step(
      self,
      signature_key: str,
      data: _SignatureInput,
      model_recipe_manager: recipe_manager.RecipeManager,
      cache_output: bool,
      pbar: progress_utils.ProgressBar,
  ) -> None:
    """Performs a single calibration step for a data sample.

    Args:
      signature_key: The signature key for the data sample.
      data: The input data for the signature.
      model_recipe_manager: The recipe manager that contains the quantization
        recipes.
      cache_output: Whether to cache the output of the model during calibration.
      pbar: The progress bar to update.
    """
    raise NotImplementedError("Subclasses must implement _calibrate_step().")

  def calibrate(
      self,
      calibration_dataset: dict[str, Iterable[_SignatureInput]],
      model_recipe_manager: recipe_manager.RecipeManager,
      cache_output: bool = False,
  ) -> None:
    """Calibrates the model."""
    total_ops = self._get_total_operations(calibration_dataset)
    with progress_utils.ProgressBar(
        total_steps=total_ops,
        description="Running Calibration:",
        disappear_on_finish=True,
    ) as pbar:
      for signature_key, dataset in calibration_dataset.items():
        for data in dataset:
          self._metadata["num_samples_calibrated"] += 1
          self._calibrate_step(
              signature_key, data, model_recipe_manager, cache_output, pbar
          )
        self._tfl_interpreter.reset_all_variables()

  def get_model_qsvs(self) -> dict[str, qtyping.QSV]:
    """Get the model qsvs.

    Returns:
      A dictionary of tensor name to QSV.
    """
    return self._model_qsvs

  def get_cached_output(self) -> list[_SignatureOutput]:
    """Get the cached output of the model."""
    return self._cached_output

  def clear_cached_output(self) -> None:
    """Clear the cached output of the model."""
    self._cached_output = []

  def reset_model_qsvs(self) -> None:
    """Reset the model qsvs."""
    self._model_qsvs = {}
    self._metadata = {"num_samples_calibrated": 0}

  def load_model_qsvs(
      self, model_qsvs: Union[str, dict[str, qtyping.QSV]]
  ) -> None:
    """Load the model qsvs.

    Args:
      model_qsvs: A dictionary of tensor name to QSV or a path to a JSON file
        that contains the model qsvs (i.e., from save_calibration_result).
    """

    if isinstance(model_qsvs, str):
      self._model_qsvs, self._metadata = (
          calibration_utils.load_calibration_results(model_qsvs)
      )
      self._metadata["num_samples_calibrated"] = self._metadata.get(
          "num_samples_calibrated", 0
      )
    else:
      self._model_qsvs = copy.deepcopy(model_qsvs)

  def save_calibration_result(
      self, file_path: str, extra_metadata: Mapping[str, str] | None = None
  ) -> None:
    """Saves the calibration result to a json file.

    Args:
      file_path: Path to save the calibration result.
      extra_metadata: Extra metadata to save.
    """
    with open(file_path, "w") as f:
      json.dump(
          {
              "model_qsvs": self._model_qsvs,
              "metadata": {**self._metadata, **(extra_metadata or {})},
          },
          f,
          cls=calibration_utils.NumpyEncoder,
      )

  def get_signature_list(self) -> list[str]:
    """Get the signature list of the model."""
    return self._tfl_interpreter.get_signature_list()

  def _update_qsvs(
      self,
      op_qsvs: dict[str, qtyping.QSV],
      ignore_tensor_names: set[str],
      qsv_update_func: Callable[[qtyping.QSV, qtyping.QSV], qtyping.QSV],
  ) -> set[str]:
    """Update the model qsvs with the new values.

    Args:
      op_qsvs: A dictionary of tensor name to QSV.
      ignore_tensor_names: A set of tensor names to ignore.
      qsv_update_func: The function to update the QSVs across calibration steps.

    Returns:
      A set of tensor names that are updated.
    """
    updated_tensor_names = set()
    for tensor_name, qsv in op_qsvs.items():
      if tensor_name in ignore_tensor_names:
        continue
      if tensor_name not in self._model_qsvs:
        self._model_qsvs[tensor_name] = qsv
      else:
        updated_qsv = qsv_update_func(self._model_qsvs[tensor_name], qsv)
        self._model_qsvs[tensor_name] = updated_qsv
      updated_tensor_names.add(tensor_name)
    return updated_tensor_names

  def _get_total_operations(self, calibration_dataset) -> int:
    """Get the total number of OPs to go through while calibrating the model."""
    data_sizes = {
        key: len(list(dataset)) for key, dataset in calibration_dataset.items()
    }
    total_ops = 0
    for key, value in data_sizes.items():
      subgraph_idx = tfl_interpreter_utils.get_signature_main_subgraph_index(
          self._tfl_interpreter, key
      )
      subgraph_inds = [subgraph_idx]
      total_operators = 0
      while subgraph_inds:
        subgraph_idx = subgraph_inds.pop()
        subgraph = self._flatbuffer_model.subgraphs[subgraph_idx]
        subgraph_operators = subgraph.operators
        if not any(
            isinstance(op, qtyping.IOOperator) for op in subgraph.operators
        ):
          subgraph_operators += (
              tfl_flatbuffer_utils.get_subgraph_input_output_operators(subgraph)
          )
        total_operators += len(subgraph_operators)
        for op in subgraph_operators:
          subgraph_inds.extend(
              tfl_flatbuffer_utils.get_op_side_effect_subgraphs(op)
          )
      total_ops += value * total_operators
    return total_ops


class _InferenceOnlyCalibrator(Calibrator):
  """Calibrator that only supports inference and not calibration."""

  @override
  def _create_interpreter(
      self,
      float_tflite: Union[str, bytes],
      num_threads: int,
  ) -> tfl.Interpreter:
    return tfl_interpreter_utils.create_tfl_interpreter(
        float_tflite,
        use_xnnpack=True,
        num_threads=num_threads,
        preserve_all_tensors=False,
    )

  @override
  def _calibrate_step(
      self,
      signature_key: str,
      data: _SignatureInput,
      model_recipe_manager: recipe_manager.RecipeManager,
      cache_output: bool,
      pbar: progress_utils.ProgressBar,
  ) -> None:
    raise NotImplementedError(
        "InferenceOnlyCalibrator does not support calibration."
    )


class _PreserveAllTensorsCalibrator(Calibrator):
  """Calibrator using the preserve_all_tensors interpreter mode."""

  @override
  def _create_interpreter(
      self,
      float_tflite: Union[str, bytes],
      num_threads: int,
  ) -> tfl.Interpreter:
    return tfl_interpreter_utils.create_tfl_interpreter(
        float_tflite,
        use_xnnpack=True,
        num_threads=num_threads,
        preserve_all_tensors=True,
    )

  @override
  def _calibrate_step(
      self,
      signature_key: str,
      data: _SignatureInput,
      model_recipe_manager: recipe_manager.RecipeManager,
      cache_output: bool,
      pbar: progress_utils.ProgressBar,
  ) -> None:
    op_codes = self._flatbuffer_model.operatorCodes
    # Step0: get subgraph index.
    subgraph_idx = tfl_interpreter_utils.get_signature_main_subgraph_index(
        self._tfl_interpreter, signature_key
    )
    # Initialize tensor names updated in this round of calibration.
    updated_tensor_names = set()

    # Step1: run tfl interpreter on subgraph to get tensor content.
    signature_output = tfl_interpreter_utils.invoke_interpreter_signature(
        self._tfl_interpreter, data, signature_key
    )
    if cache_output:
      self._cached_output.append(signature_output)

    # Step2: go through each op in subgraph to update quantization
    # statistic values.
    subgraphs_inds = [subgraph_idx]
    while subgraphs_inds:
      subgraph_ind = subgraphs_inds.pop()
      self._tensor_content_map.update(
          tfl_interpreter_utils.get_tensor_name_to_content_map(
              self._tfl_interpreter, subgraph_ind
          )
      )
      subgraph = self._flatbuffer_model.subgraphs[subgraph_ind]
      graph_info = qtyping.GraphInfo(
          subgraph.tensors, self._flatbuffer_model.buffers
      )
      # Add input/output operators if they are not in the subgraph.
      if not any(
          isinstance(op, qtyping.IOOperator) for op in subgraph.operators
      ):
        subgraph.operators += (
            tfl_flatbuffer_utils.get_subgraph_input_output_operators(subgraph)
        )
      for op in subgraph.operators:
        pbar.update_single_step()
        if isinstance(op, qtyping.IOOperator):
          op_key = op.op_key
        else:
          op_code = op_codes[op.opcodeIndex].builtinCode
          if op_code not in tfl_flatbuffer_utils.TFL_OP_CODE_TO_NAME:
            continue
          op_key = tfl_flatbuffer_utils.TFL_OP_CODE_TO_NAME[op_code]
        # Step2.1: query the quantization_recipe to get op quantization
        # settings.
        op_scope = tfl_flatbuffer_utils.get_op_scope(op, subgraph.tensors)
        algorithm_name, _ = model_recipe_manager.get_quantization_configs(
            op_key, op_scope
        )
        if algorithm_name == algorithm_manager.AlgorithmName.NO_QUANTIZE:
          continue
        if policy.is_non_quantizable_composite_op(op):
          continue

        # Step2.2: query algorithm_manager to get/call the related
        # calibration function.
        calibrate_func = algorithm_manager.get_quantization_func(
            algorithm_name, op_key, qtyping.QuantizeMode.CALIBRATE
        )
        op_qsvs = calibrate_func(op, graph_info, self._tensor_content_map)
        # Step3: Update tensor qsvs with the new values. Ignore the tensor
        # names that are already updated in this round of calibration.
        if not self._is_custom_qsv_update_func:
          qsvs_update_func = algorithm_manager.get_update_qsv_func(
              algorithm_name, op_key
          )
        else:
          qsvs_update_func = self._qsv_update_func
        op_updated_tensor_name = self._update_qsvs(
            op_qsvs, updated_tensor_names, qsvs_update_func
        )
        updated_tensor_names.update(op_updated_tensor_name)

        # Step4: Invoke any subgraphs invoked as a side effect of the op.
        subgraphs_inds.extend(
            tfl_flatbuffer_utils.get_op_side_effect_subgraphs(op)
        )



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import argparse
import pathlib
import tempfile

from absl.testing import absltest
from absl.testing import parameterized

from ai_edge_litert.tools import mmap_utils
from ai_edge_quantizer import aeq
from ai_edge_quantizer import model_validator
from ai_edge_quantizer import qtyping
from ai_edge_quantizer.utils import litertlm_utils
from ai_edge_quantizer.utils import test_utils
from ai_edge_quantizer.utils import tfl_interpreter_utils
from ai_edge_quantizer.utils import validation_utils

TEST_DATA_PREFIX_PATH = test_utils.get_path_to_datafile(".")
RECIPE_PREFIX_PATH = pathlib.Path(TEST_DATA_PREFIX_PATH) / "recipes"


class AeqTest(parameterized.TestCase):

  def _validate_quantized_model(
      self,
      float_model_buffer: qtyping.BufferType,
      quantized_model_buffer: qtyping.BufferType,
  ):
    # Create some random test data.
    test_data = tfl_interpreter_utils.create_random_normal_input_data(
        float_model_buffer, num_samples=10
    )

    # Run the evaluation.
    result = model_validator.compare_model(
        float_model_buffer,
        quantized_model_buffer,
        test_data,
        error_metric="mse",
        compare_fn=validation_utils.get_validation_func("mse"),
        validate_output_tensors_only=True,
    )

    # Verify that the results are indeed reasonable.
    for mse in result.get_all_tensor_results().values():
      self.assertLess(mse, 1e-4)

  @parameterized.named_parameters(
      (
          "from_recipe_file",
          str(RECIPE_PREFIX_PATH / "dynamic_wi8_afp32_recipe.json"),
      ),
      ("from_recipe_name", "dynamic_wi8_afp32"),
  )
  def test_quantize_and_validate_single_tflite_file(self, recipe_file: str):
    model_file = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / "tests/models/conv_fc_mnist.tflite"
    )

    with tempfile.TemporaryDirectory() as output_dir:
      self.assertEqual(
          aeq.main(
              argparse.Namespace(
                  model_file=model_file,
                  recipe=recipe_file,
                  output_dir=output_dir,
                  overwrite_outputs=False,
              )
          ),
          0,
      )

      model_basename = pathlib.Path(model_file).stem
      recipe_basename = pathlib.Path(recipe_file).stem
      output_filename = f"{model_basename}_{recipe_basename}.tflite"
      output_path = str(pathlib.Path(output_dir) / output_filename)

      self.assertTrue(pathlib.Path(output_path).exists())

      original_size = pathlib.Path(model_file).stat().st_size
      quantized_size = pathlib.Path(output_path).stat().st_size

      # Check whether the file size of the quantized model is roughly 4X smaller
      # than that of the float model
      self.assertLess(quantized_size, original_size * 0.35)

      # Validate the quantized model.
      self._validate_quantized_model(
          mmap_utils.get_file_contents(model_file),
          mmap_utils.get_file_contents(output_path),
      )

  @parameterized.named_parameters(
      (
          "recipe_mapping_path",
          str(RECIPE_PREFIX_PATH / "dynamic_wi8_afp32_litertlm_recipe.json"),
      ),
      (
          "default_recipe_path",
          str(RECIPE_PREFIX_PATH / "dynamic_wi8_afp32_recipe.json"),
      ),
      ("default_recipe_name", "dynamic_wi8_afp32"),
  )
  def test_quantize_and_validate_tflite_models_in_litertlm_file(
      self, recipe: str
  ):
    model_file = str(
        pathlib.Path(TEST_DATA_PREFIX_PATH)
        / "tests/models/conv_fc_mnist.litertlm"
    )

    with tempfile.TemporaryDirectory() as output_dir:
      self.assertEqual(
          aeq.main(
              argparse.Namespace(
                  model_file=model_file,
                  recipe=recipe,
                  output_dir=output_dir,
                  overwrite_outputs=False,
              )
          ),
          0,
      )

      model_basename = pathlib.Path(model_file).stem
      recipe_basename = pathlib.Path(recipe).stem
      output_filename = f"{model_basename}_{recipe_basename}.litertlm"
      output_path = str(pathlib.Path(output_dir) / output_filename)

      self.assertTrue(pathlib.Path(output_path).exists())

      original_size = pathlib.Path(model_file).stat().st_size
      quantized_size = pathlib.Path(output_path).stat().st_size

      # Check whether the file size of the quantized model is roughly 4X smaller
      # than that of the float model (keep in mind that the LiteRT-LM header
      # size did not change).
      self.assertLess(quantized_size, original_size * 0.37)

      # Check whether the system metadata match, modulo the `uuid` and
      # `timestamp` fields, which should not..
      original_litertlm = litertlm_utils.LiteRTLMFile(model_file)
      quantized_litertlm = litertlm_utils.LiteRTLMFile(output_path)
      original_system_metadata = original_litertlm.get_system_metadata()
      quantized_system_metadata = quantized_litertlm.get_system_metadata()
      for tag in ("uuid", "creation_timestamp"):
        self.assertNotEqual(
            original_system_metadata.pop(tag),
            quantized_system_metadata.pop(tag),
        )
      self.assertEqual(original_system_metadata, quantized_system_metadata)

      # Check whether the section metadata match.
      self.assertEqual(
          original_litertlm.get_section_metadata(0),
          quantized_litertlm.get_section_metadata(0),
      )

      # Validate the quantized model.
      self._validate_quantized_model(
          original_litertlm.get_section_buffer(0),
          quantized_litertlm.get_section_buffer(0),
      )


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

import io
import sys
from unittest import mock

from absl.testing import absltest
from absl.testing import parameterized
import tqdm

import os
import io
from ai_edge_quantizer.utils import progress_utils
from ai_edge_quantizer.utils import test_utils


class ProgressBarTest(absltest.TestCase):

  def setUp(self):
    super().setUp()
    self.mock_tqdm = self.enter_context(
        mock.patch.object(tqdm, 'tqdm', autospec=True, spec_set=True)
    )

  def test_progress_bar_update(self):
    mock_progress_bar_instance = self.mock_tqdm.return_value
    with progress_utils.ProgressBar(total_steps=10, enable=True) as pb:
      pb.update_single_step()
      pb.update_single_step()

    self.mock_tqdm.assert_called_once_with(
        total=10, desc='', leave=True, disable=False
    )
    self.assertEqual(mock_progress_bar_instance.update.call_count, 2)
    mock_progress_bar_instance.update.assert_called_with(1)
    mock_progress_bar_instance.close.assert_called_once()

  def test_progress_bar_disable(self):
    mock_progress_bar_instance = self.mock_tqdm.return_value
    with progress_utils.ProgressBar(total_steps=10, enable=False):
      pass
    self.mock_tqdm.assert_called_once_with(
        total=10, desc='', leave=True, disable=True
    )
    mock_progress_bar_instance.close.assert_called_once()

  def test_progress_bar_disappear_on_finish(self):
    mock_progress_bar_instance = self.mock_tqdm.return_value
    with progress_utils.ProgressBar(
        total_steps=10, disappear_on_finish=True, enable=True
    ):
      pass
    self.mock_tqdm.assert_called_once_with(
        total=10, desc='', leave=False, disable=False
    )
    mock_progress_bar_instance.close.assert_called_once()


class ProgressReportTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.mock_time = self.enter_context(
        mock.patch.object(progress_utils, 'time', autospec=True, spec_set=True)
    )
    self.mock_tracemalloc = self.enter_context(
        mock.patch.object(
            progress_utils, 'tracemalloc', autospec=True, spec_set=True
        )
    )

  @parameterized.named_parameters(
      ('trace_memory_enabled', True), ('trace_memory_disabled', False)
  )
  def test_generate_progress_report(self, trace_memory: bool):
    self.mock_time.time.side_effect = [100.0, 105.5]  # Start time, end time.

    if trace_memory:
      self.mock_tracemalloc.is_tracing.return_value = False
      self.mock_tracemalloc.start.side_effect = None
      self.mock_tracemalloc.stop.return_value = None
      self.mock_tracemalloc.get_traced_memory.return_value = (
          1 * 1024 * 1024,
          2 * 1024 * 1024,
      )

    progress_report = progress_utils.ProgressReport(trace_memory=trace_memory)
    progress_report.capture_progess_start()

    original_model_path = test_utils.get_path_to_datafile(
        '../tests/models/conv_fc_mnist.tflite'
    )
    quantized_model_path = test_utils.get_path_to_datafile(
        '../tests/models/mnist_quantized.tflite'
    )
    with open(original_model_path, 'rb') as f:
      original_model_data = f.read()
    with open(quantized_model_path, 'rb') as f:
      quantized_model = f.read()

    mock_stdout = io.StringIO()
    with mock.patch.object(sys, 'stdout', mock_stdout):
      progress_report.generate_progress_report(
          len(original_model_data), len(quantized_model)
      )

    output = mock_stdout.getvalue()
    self.assertIn('Original model size: 200.23 KiB', output)
    self.assertIn('Quantized model size: 53.08 KiB', output)
    self.assertIn('Quantization Ratio: 0.27', output)
    self.assertIn('Total time: 5.50 s', output)

    if trace_memory:
      self.mock_tracemalloc.is_tracing.assert_called_once_with()
      self.mock_tracemalloc.start.assert_called_once_with()
      self.mock_tracemalloc.stop.assert_called_once_with()
      self.assertIn('Memory peak: 2.00 MiB', output)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Utilities to read named recipe files."""

from collections.abc import Callable, Collection
import copy
import inspect
import json
import logging
import pathlib
from typing import TypeGuard

import os
import io
from ai_edge_quantizer import qtyping
from ai_edge_quantizer import recipe as recipes

UnresolvedQuantRecipeMapping = dict[str, qtyping.ModelQuantizationRecipe | str]
QuantRecipeMapping = dict[str, qtyping.ModelQuantizationRecipe]
QuantRecipeFileContents = (
    qtyping.ModelQuantizationRecipe | UnresolvedQuantRecipeMapping
)


# Create a mapping of recipe names to parsed JSON files.
_RECIPE_REPO_PATH = pathlib.Path(__file__).parent / '../recipes'
# _NAMED_RECIPES: dict[str, QuantRecipeFileContents] | None = None
_NAMED_RECIPES: dict[str, Callable[..., QuantRecipeFileContents]] | None = None


def _get_recipe_from_path(
    path: qtyping.Path, extra_dirs: Collection[qtyping.Path] | None = None
) -> QuantRecipeFileContents:
  """Loads a JSON file from an absolute or relative path."""
  path = pathlib.Path(path)

  # Check whether the absolute or relative path exists.
  if (
      os.path.exists(  # Assume this is an absolute or relative (to cwd) path.
          actual_path := path
      )
      or os.path.exists(  # Assume it is relative to the recipes directory.
          actual_path := _RECIPE_REPO_PATH / path
      )
      or (
          extra_dirs
          and any(
              os.path.exists(actual_path := pathlib.Path(dir) / path)
              for dir in extra_dirs
          )
      )
  ):
    logging.info('Loading recipe from %s.', actual_path)
    with open(actual_path, 'r') as f:
      return json.load(f)

  raise ValueError(f'Failed to load/resolve recipe "{path}".')


def _get_named_recipe(
    name: str, default: QuantRecipeFileContents | None = None
) -> QuantRecipeFileContents | None:
  """Looks up a recipe from the `recipes/` directory by its name."""
  global _NAMED_RECIPES

  # Populate the named recipes if this has not been done yet.
  if _NAMED_RECIPES is None:
    # Initialize the named recipes with the functions in the `recipe` module.
    _NAMED_RECIPES = dict(inspect.getmembers(recipes, inspect.isfunction))

    # Look for additional named recipes in the `recipes/` directory.
    for recipe_path in _RECIPE_REPO_PATH.iterdir():
      if (recipe_name := recipe_path.name).endswith('.json'):
        recipe_name = recipe_name.removesuffix('.json')
        recipe_name = recipe_name.removesuffix('_recipe')
        if recipe_name not in _NAMED_RECIPES:
          with open(recipe_path, 'r') as f:
            recipe = json.load(f)
          _NAMED_RECIPES[recipe_name] = lambda recipe=recipe: copy.deepcopy(
              recipe
          )

    logging.info('Loaded named recipes: %s', _NAMED_RECIPES.keys())

  if recipe_fun := _NAMED_RECIPES.get(name):
    return recipe_fun()
  return default


def resolve_recipe(
    recipe_name_or_path: qtyping.Path,
    extra_dirs: Collection[qtyping.Path] | None = None,
) -> QuantRecipeFileContents:
  """Loads a quantization recipe given by name or path.

  Recipes can be identified either:

    1. By a canonical name, where `name` corresponds to a file
      `recipes/name_recipe.json` in the code's base directory,
    2. By absolute path,
    3. By relative path,
    4. By a file name relative to the code's `recipes/` directory,
    5. By a file name relative to any of the directories in `extra_dirs`,

  in the above order.

  Args:
    recipe_name_or_path: The name of, or path to, a recipe file.
    extra_dirs: An optional collection of extra directories in which to look for
      `recipe_name_or_path`.

  Returns:
    Either a:
      * `qtyping.ModelQuantizationRecipe` if the name or path refered to a
        quantization recipe, or a
      * `dict[str, qtyping.ModelQuantizationRecipe | str]` if the name or path
        refered to a LiteRT-LM mapping from `model_type` to quantization recipe.

  Raises:
    ValueError if the recipe name or path could not be resolved.
  """
  # Check whether this is a named recipe.
  if recipe := _get_named_recipe(recipe_name_or_path):
    logging.info(
        'Loading named recipe "%s" from %s.',
        recipe_name_or_path,
        _RECIPE_REPO_PATH,
    )
    return recipe

  # Otherwise, assume we were given a file name/path.
  return _get_recipe_from_path(recipe_name_or_path, extra_dirs)


def _is_quant_recipe(
    recipe: QuantRecipeFileContents,
) -> TypeGuard[qtyping.ModelQuantizationRecipe]:
  return isinstance(recipe, list) and all(
      isinstance(val, dict) and all(isinstance(key, str) for key in val)
      for val in recipe
  )


def _as_quant_recipe_or_none(
    recipe: QuantRecipeFileContents,
) -> qtyping.ModelQuantizationRecipe | None:
  if _is_quant_recipe(recipe):
    return recipe
  else:
    return None


def _is_unresolved_recipe_mapping(
    recipe: QuantRecipeFileContents,
) -> TypeGuard[UnresolvedQuantRecipeMapping]:
  return isinstance(recipe, dict) and all(
      isinstance(key, str)
      and (isinstance(value, str) or _is_quant_recipe(value))
      for key, value in recipe.items()
  )


def resolve_litertlm_recipe_or_mapping(
    recipe_name_or_path: qtyping.Path,
) -> QuantRecipeMapping:
  """Loads a LiteRT-LM quantization recipe given by name or path.

  The named recipe or JSON file is either:
  1. A quantization recipe, which will be used for all models in the LiteRT-LM
    file.
  2. A mapping of strings, corresponding to a `model_type` or `'default'`, to
    either:
    1. A `list[dict[str, Any]]` corresponding to a quantization recipe for the
      `model_type`,
    2. A canonical name, where `name` corresponds to a file
      `recipes/name_recipe.json` in the code's base directory,
    3. A `str` containing the absolute path to a JSON file containing the
      quantization recipe for the `model_type`,
    4. A `str` containing the  path, relative to the current working directory,
      to a JSON file containing the quantization recipe for the `model_type`,
    5. A `str` containing the  path, relative to the directory containing
      `recipe_mapping_path`, to a JSON file containing the quantization recipe
      for the `model_type`.

  Args:
    recipe_name_or_path: The name of, or path to, a JSON file containing the
      default recipe or a per-`model_type` recipe mapping.

  Returns:
    A `QuantRecipeMapping` mapping `model_type` or `'default'` to a quantization
    recipe.

  Raises:
    ValueError if the recipe name or path could not be resolved.
  """
  # Load the recipe or per model_type dictionary of recipes.
  recipe_or_mapping: QuantRecipeFileContents = resolve_recipe(
      recipe_name_or_path
  )

  # If we got a quant recipe, then return a mapping with it set as the default.
  if recipe := _as_quant_recipe_or_none(recipe_or_mapping):
    logging.info(
        'Using recipe "%s" as the default recipe for all models in the'
        ' LiteRT-LM file.',
        recipe_name_or_path,
    )
    return {'default': recipe}

  if not _is_unresolved_recipe_mapping(recipe_or_mapping):
    raise ValueError(
        f'Expected LiteRT-LM recipe {recipe_name_or_path} to contain a mapping'
        ' of `model_type` to recipes, recipe names, or recipe file paths.'
    )

  # Loop over the map of model_type to recipe and resolve any names/paths.
  extra_dirs = [pathlib.Path(recipe_name_or_path).parent]
  recipe_mapping: QuantRecipeMapping = {}
  for model_type, recipe_or_path in recipe_or_mapping.items():
    if (recipe := _as_quant_recipe_or_none(recipe_or_path)) is None:
      recipe: QuantRecipeFileContents = resolve_recipe(
          recipe_or_path, extra_dirs
      )
      if not _is_quant_recipe(recipe):
        raise ValueError(
            f'Expected LiteRT-LM recipe {recipe_name_or_path} to contain a'
            f' recipe name or recipe file path for model_type {model_type}, but'
            f' got object of type {type(recipe)} instead.'
        )
    recipe_mapping[model_type] = recipe

  return recipe_mapping



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

from ai_edge_quantizer.utils import qsv_utils


class QsvUtilsTest(parameterized.TestCase):

  @parameterized.named_parameters(
      dict(
          testcase_name="zero_smoothing_factor",
          smoothing_factor=0,
          expected_vals={"min": -1000.0, "max": 800.0},
      ),
      dict(
          testcase_name="one_smoothing_factor",
          smoothing_factor=1,
          expected_vals={"min": -10.0, "max": 8.0},
      ),
      dict(
          testcase_name="normal_smoothing_factor",
          smoothing_factor=0.99,
          expected_vals={"min": -19.9, "max": 15.92},
      ),
  )
  def test_moving_average_update(self, smoothing_factor, expected_vals):
    old_qsv = {"min": -10.0, "max": 8.0}
    # Large values to mimic outlier.
    new_qsv = {"min": -1000.0, "max": 800.0}
    updated_qsv = qsv_utils.moving_average_update(
        old_qsv, new_qsv, smoothing_factor=smoothing_factor
    )
    self.assertAlmostEqual(updated_qsv["min"], expected_vals["min"])
    self.assertAlmostEqual(updated_qsv["max"], expected_vals["max"])

  def test_moving_average_update_with_array(self):
    old_qsv = {"min": np.array([-10.0, -20.0]), "max": np.array([8.0, 10.0])}
    new_qsv = {
        "min": np.array([-1000.0, -2000.0]),
        "max": np.array([800.0, 1000.0]),
    }
    updated_qsv = qsv_utils.moving_average_update(
        old_qsv, new_qsv, smoothing_factor=0.99
    )
    np.testing.assert_array_almost_equal(
        updated_qsv["min"], np.array([-19.9, -39.8])
    )
    np.testing.assert_array_almost_equal(
        updated_qsv["max"], np.array([15.92, 19.9])
    )

  def test_moving_average_update_no_old_qsv(self):
    new_qsv = {"min": -1000.0, "max": 800.0}
    updated_qsv = qsv_utils.moving_average_update(None, new_qsv)
    self.assertEqual(updated_qsv, new_qsv)

  @parameterized.named_parameters(
      dict(
          testcase_name="scalar",
          old_qsv={"min": -10.0, "max": 8.0},
          new_qsv={"min": -1000.0, "max": 1.0},
          expected_qsv={"min": -1000.0, "max": 8.0},
      ),
      dict(
          testcase_name="2darray",
          old_qsv={
              "min": np.array([[-19.0], [20.0]]),
              "max": np.array([[21.0], [250.0]]),
          },
          new_qsv={
              "min": np.array([[-1000.0], [25.0]]),
              "max": np.array([[33.0], [100.0]]),
          },
          expected_qsv={
              "min": np.array([[-1000.0], [20.0]]),
              "max": np.array([[33.0], [250.0]]),
          },
      ),
  )
  def test_min_max_update(self, old_qsv, new_qsv, expected_qsv):
    updated_qsv = qsv_utils.min_max_update(old_qsv, new_qsv)
    np.testing.assert_array_equal(updated_qsv["min"], expected_qsv["min"])
    np.testing.assert_array_equal(updated_qsv["max"], expected_qsv["max"])

  def test_min_max_update_no_old_qsv(self):
    new_qsv = {"min": -1000.0, "max": 800.0}
    updated_qsv = qsv_utils.min_max_update(None, new_qsv)
    self.assertEqual(updated_qsv, new_qsv)

  def test_gptq_and_moving_average_update_no_old_qsv(self):
    new_qsv = {"min": -1000.0, "max": 800.0}
    updated_qsv = qsv_utils.gptq_and_moving_average_update(None, new_qsv)
    self.assertEqual(updated_qsv, new_qsv)

  def test_gptq_and_moving_average_update(self):
    qsv = {
        "min": 1.0,
        "max": 10.0,
        "hessian": np.array([[1.0, 0.0], [0.0, 1.0]]),
        "num_samples": 10,
    }
    new_qsv = {
        "min": 0.0,
        "max": 12.0,
        "hessian": np.array([[3.0, 0.0], [0.0, 3.0]]),
        "num_samples": 10,
    }
    updated_qsv = qsv_utils.gptq_and_moving_average_update(qsv, new_qsv)

    # moving_average_update check with smoothing_factor=0.95
    expected_min = 0.95 * 1.0 + 0.05 * 0.0
    expected_max = 0.95 * 10.0 + 0.05 * 12.0
    self.assertAlmostEqual(updated_qsv["min"], expected_min)
    self.assertAlmostEqual(updated_qsv["max"], expected_max)

    # gptq merge hessian check
    expected_hessian = (10 * qsv["hessian"] + 10 * new_qsv["hessian"]) / 20
    np.testing.assert_array_almost_equal(
        updated_qsv["hessian"], expected_hessian
    )
    self.assertEqual(updated_qsv["num_samples"], 20)

  @parameterized.named_parameters(
      dict(
          testcase_name="zero_samples_old",
          old_ns=0,
          new_ns=10,
          expected_h=np.array([[3.0, 0.0], [0.0, 3.0]]),
          expected_ns=10,
      ),
      dict(
          testcase_name="zero_samples_new",
          old_ns=10,
          new_ns=0,
          expected_h=np.array([[1.0, 0.0], [0.0, 1.0]]),
          expected_ns=10,
      ),
      dict(
          testcase_name="zero_samples_both",
          old_ns=0,
          new_ns=0,
          expected_h=np.array([[3.0, 0.0], [0.0, 3.0]]),
          expected_ns=0,
      ),
  )
  def test_gptq_and_moving_average_update_zero_samples(
      self, old_ns, new_ns, expected_h, expected_ns
  ):
    qsv = {
        "min": 1.0,
        "max": 10.0,
        "hessian": np.array([[1.0, 0.0], [0.0, 1.0]]),
        "num_samples": old_ns,
    }
    new_qsv = {
        "min": 0.0,
        "max": 12.0,
        "hessian": np.array([[3.0, 0.0], [0.0, 3.0]]),
        "num_samples": new_ns,
    }
    updated_qsv = qsv_utils.gptq_and_moving_average_update(qsv, new_qsv)

    np.testing.assert_array_almost_equal(updated_qsv["hessian"], expected_h)
    self.assertEqual(updated_qsv["num_samples"], expected_ns)


if __name__ == "__main__":
  absltest.main()



# Copyright 2024 The AI Edge Quantizer Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Utility functions to display a progress bar and progress report."""

from collections.abc import Sequence
import logging
import time
import tracemalloc

import tqdm


def _format_base(
    value: int | float, multiple: int | float, units: Sequence[str]
) -> str:
  """Converts `value` to a string using the given `units` per `multiple`.

  Creates a `str` representation of `value / (multiple ** k)` with `units[k]`
  appended, where `k` is the largest integer in the range `[0, len(units))`
  such that `multiple ** k` is less than `value`, rounded to two decimal digits.

  For all `value` in the range `[1, multiple**len(units))`, the represented
  value will be in the range `[1, multiple)`.

  E.g. if `multiple=10` and `units=['mm', 'cm', 'dm', 'm']`, the following
  `value`s will generate the following outputs:
  * `1`: `'1.00 mm'`,
  * `9`: `'9.00 mm'`,
  * `10`: `'1.00 cm'`,
  * `110`: `'1.10 dm'`,
  * `1002`: `'1.00 m'`.

  Args:
    value: The numerical value to convert to a `str`.
    multiple: The multiple used for the `units`.
    units: A sequence of unit names.

  Returns:
    A `str` representation of an appropriately scaled `value` with the given
    `units` appended.
  """
  for unit in units[:-1]:
    if abs(value) < multiple:
      return f'{value:.2f} {unit}'
    value /= multiple
  return f'{value:.2f} {units[-1]}'


def _format_bytes(value: int | float) -> str:
  return _format_base(value, 1024, ['B', 'KiB', 'MiB', 'GiB'])


def _format_ns(value: int | float) -> str:
  return _format_base(value, 1000, ['ns', 'us', 'ms', 's'])


class ProgressBar:
  """A Progress Bar that can be used to track the progress of a process."""

  def __init__(
      self,
      total_steps: int,
      description: str = '',
      disappear_on_finish: bool = False,
      enable: bool | None = None,
  ):

    if enable is None:
      # Progress bar will be skipped for smaller models.
      disable = total_steps < 1000
    else:
      disable = not enable
    self._progress_bar = tqdm.tqdm(
        total=total_steps,
        desc=description,
        leave=not disappear_on_finish,
        disable=disable,
    )

  def __enter__(self):
    """Enters the context manager."""
    return self

  def __exit__(self, exc_type, exc_value, traceback):
    """Exits the context manager and closes the progress bar."""
    self.close()

  def update_single_step(self):
    """Updates the progress bar by a single step."""
    self._progress_bar.update(1)

  def close(self):
    """Closes the progress bar."""
    self._progress_bar.close()


class ProgressReport:
  """A class to generate a progress report for the quantization process.

  If initialized with `trace_memory=True`, it will also track the peak memory
  use using `tracemalloc`, which may hurt performance and interfere with other
  code using `tracemalloc` concurrently.
  """

  _model_name: str | None
  _trace_memory: bool
  _start_time: float | None = None
  _tracemalloc_started_by_me: bool = False

  def __init__(self, model_name: str | None = None, trace_memory: bool = False):
    self._model_name = model_name
    self._trace_memory = trace_memory

  def capture_progess_start(self):
    self._start_time = time.time()
    if self._trace_memory:
      logging.warning(
          'Progress bar reporting with `trace_memory=True` is enabled which may'
          ' significantly slow down your computations!'
      )
      self._tracemalloc_started_by_me = not tracemalloc.is_tracing()
      if self._tracemalloc_started_by_me:
        tracemalloc.start()
      else:
        tracemalloc.reset_peak()

  def _capture_progress_end(self) -> int | None:
    if self._trace_memory:
      _, mem_peak_bytes = tracemalloc.get_traced_memory()
      if self._tracemalloc_started_by_me:
        tracemalloc.stop()
        self._tracemalloc_started_by_me = False
      return mem_peak_bytes

  def generate_progress_report(
      self, original_model_size: int, quantized_model_size: int
  ):
    """Prints out the progress report."""
    # Collect the metrics.
    quantization_ratio = quantized_model_size / original_model_size
    total_time = time.time() - self._start_time
    mem_peak_bytes = self._capture_progress_end()

    # Print out the progress report.
    if self._model_name:
      print(f'Model name: {self._model_name}')
    print(f'Original model size: {_format_bytes(original_model_size)}')
    print(f'Quantized model size: {_format_bytes(quantized_model_size)}')
    print(
        f'Quantization Ratio: {quantization_ratio:.2f}'
        f' ({1/quantization_ratio:.1f}x smaller)'
    )
    print(f'Total time: {_format_ns(total_time * 1e9)}')
    if mem_peak_bytes is not None:
      print(f'Memory peak: {_format_bytes(mem_peak_bytes)}')
